// Copyright (c) ONNX Project Contributors.
// Licensed under the MIT license.

#pragma once

namespace ONNX_NAMESPACE {

// Represents the most recent release version. Updated with every release.
constexpr const char* LAST_RELEASE_VERSION = "1.8.1";

}
